/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client mount functions
   Copyright (C) Andrew Tridgell              1994-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-1998
   Copyright (C) Volker Lendecke              1996-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef SYSLOG
#undef SYSLOG
#endif

#include "includes.h"
#include "trans2.h"

/* only the linux kernel supports smbfs at the moment */
#ifdef LINUX_MOUNT
#include "linux/smb_fs.h"
/* #define SMBFS_DEBUG */
#endif

/* value for unused fid field in trans2 secondary request */
#define FID_UNUSED (0xFFFF)

#define CNV_LANG(s) dos2unix_format(s,False)

extern file_info def_finfo;

extern int DEBUGLEVEL;

#ifdef LINUX_MOUNT

/****************************************************************************
 only the linux kernel supports the smb mount ioctls
 ****************************************************************************/


/****************************************************************************
 turn client into a daemon (?!)
 ****************************************************************************/
static void cli_mount_daemonize(void)
{
	int i;
	if ((i = fork()) < 0)
	{
		DEBUG(0, ("could not fork\n"));
	}
	if (i > 0)
	{
		/* parent simply exits */
		exit(0);
	}
	setsid();
	chdir("/");
}

/****************************************************************************
 close all local files.  leave the client fd open
 ****************************************************************************/
static void cli_mount_close_our_files(int client_fd)
{
	int i;
	for (i = 0; i < NR_OPEN; i++)
	{
		if (i != client_fd)
		{
			close(i);
		}
	}
}

/****************************************************************************
 usr1_handler
 ****************************************************************************/
static void cli_mount_usr1_handler(int x)
{
	return;
}

/****************************************************************************
 Send a login and store the connection options. 
 keeps connection options independent of linux kernel changes.
 ****************************************************************************/
static BOOL cli_mount_get_opts(struct cli_state *cli, int t_idx,
			struct smb_conn_opt *opts)
{
	/* copy connection options, ready to do an ioctl that the linux kernel supports */

	opts->protocol      = cli->protocol;

	/* Connection options */
	DEBUG(0,("cli_mount_reestablish_connection: volker, case_handling appears not to exist\n"));
	opts->case_handling = CASE_LOWER;
	opts->max_xmit      = cli->max_xmit;
	opts->server_uid    = cli->uid;
	opts->tid           = cli->con[t_idx].cnum;

	/* The following are LANMAN 1.0 options */
	opts->sec_mode       = cli->sec_mode;
	DEBUG(0,("cli_mount_reestablish_connection: volker, max_mux and max_vcs appear not to exist\n"));
#if 0
	opts->max_mux        = cli->max_mux;
	opts->max_vcs        = cli->max_vcs;
#endif
	DEBUG(0,("cli_mount_reestablish_connection: volker, can you check the rawmode parameter?\nit should be two separate read and write raw parameters\n"));
#if 0
	opts->rawmode       = cli->rawmode;
#endif
	opts->sesskey       = cli->sesskey;

	/* The following are NT LM 0.12 options */
	DEBUG(0,("cli_mount_reestablish_connection: and the max_raw one.  and the capabilities :-)\n"));
#if 0
	opts->maxraw        = cli->max_raw;
	opts->capabilities  = cli->capabilities;
#endif
	opts->serverzone    = cli->serverzone;

	return True;
}

/****************************************************************************
 Call the smbfs ioctl to install a connection socket,
 then wait for a signal to reconnect. Note that we do
 not exit after open_sockets() or send_login() errors,
 as the smbfs mount would then have no way to recover.
 ****************************************************************************/
void cli_mount_send_fs_socket(struct cli_state *cli, uint32 *t_idx,
				char *mount_point)
{
	BOOL closed = False;
	BOOL res = True;
	struct smb_conn_opt opts;

	while (True)
	{
		int mt_fd;

		if ((mt_fd = open(mount_point, O_RDONLY)) < 0)
		{
#ifdef SMBFS_DEBUG
			DEBUG(3,("cli_mount_send_fs_socket: can't open %s\n", mount_point));
#endif
			break;
		}		

		/*
		 * Call the ioctl even if we couldn't get a socket ...
		 * there's no point in making smbfs wait for a timeout.
		 */
		opts.fd = -1;

		if (res)
		{
			opts.fd = cli->fd;
		}

		res = ioctl(mt_fd, SMB_IOC_NEWCONN, &opts);

		if (res != 0)
		{
#ifdef SMBFS_DEBUG
			DEBUG(3,("cli_mount_send_fs_socket: ioctl failed, res=%d\n",res));
#endif
		}

		/* close the client connection */
		cli_shutdown(cli);

		/*
		 * Close all open files if we haven't done so yet.
		 */
#ifndef SMBFS_DEBUG
		if (!closed)
		{
			closed = True;
			cli_mount_close_our_files(cli->fd); /* close all but the client fd */
		}
#endif

		/*
		 * Wait for a signal from smbfs ...
		 */
		signal(SIGUSR1, &cli_mount_usr1_handler);
		pause();

#ifdef SMBFS_DEBUG
		DEBUG(3,("cli_mount_send_fs_socket: got signal, getting new socket\n"));
#endif

		/* re-establish the connection */
		res = cli_reestablish_connection(cli, t_idx);

		if (!res)
		{
#ifdef SMBFS_DEBUG
			DEBUG(3,("cli_mount_send_fs_socket: connection failed\n"));
#endif
		}
		else
		{
			cli_mount_get_opts(cli, *t_idx, &opts);
		}
	}

#ifdef SMBFS_DEBUG
	DEBUG(3,("cli_mount_send_fs_socket: exit\n"));
#endif

	exit(1);
}

/****************************************************************************
mount smbfs
****************************************************************************/
void cli_mount(struct cli_state *cli, int *t_idx, char *mount_pt, char *mount_cmd)
{
	DEBUG(3,("cli_mount: %s\n", mount_cmd));

	/*
	 * Create the background process before trying the mount.
	 * (We delay closing files to allow diagnostic messages.)
	 */

	cli_mount_daemonize();

	/* The parent has exited here, the child handles the connection: */
	if ((system(mount_cmd)) != 0)
	{
		DEBUG(0,("cli_mount: mount (%s) failed\n", mount_cmd));
	}
	else
	{
		cli_mount_send_fs_socket(cli, t_idx, mount_pt);
	}
}	

#else /* only the linux kernel supports the smb mount ioctls */

/****************************************************************************
mount smbfs
****************************************************************************/
void cli_mount(struct cli_state *cli, int *t_idx, char *mount_pt, char *mount_cmd)
{
	DEBUG(3,("cli_mount: cmd:%s pt:%s\n", mount_cmd, mount_pt));
	DEBUG(0,("cli_mount: failed - only works under linux at the moment.\n"));
	DEBUG(0,("please contact your o/s vendor and ask nicely about smb kernel support...\n"));
}

#endif /* LINUX_MOUNT - only the linux kernel supports the smb mount ioctls */

