/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client print functions
   Copyright (C) Andrew Tridgell              1994-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef SYSLOG
#undef SYSLOG
#endif

#include "includes.h"
#include "trans2.h"

/* only the linux kernel supports smbfs at the moment */
#ifdef LINUX_MOUNT
#include "linux/smb_fs.h"
/* #define SMBFS_DEBUG */
#endif

/* value for unused fid field in trans2 secondary request */
#define FID_UNUSED (0xFFFF)

#define CNV_LANG(s) dos2unix_format(s,False)

extern file_info def_finfo;

extern int DEBUGLEVEL;

/****************************************************************************
  print a file
  ****************************************************************************/
BOOL cli_print(struct cli_state *cli, int t_idx, struct client_info *info,
				FILE *f, char *lname, char *rname)
{
	uint16 fnum;
	uint32 nread=0;
	char *p;

	/* yes, you could potentially pass stdin as the input file */
	if (!f) f = fopen(lname,"r");

	if (!f)
	{
		DEBUG(0,("Error opening local file %s\n",lname));
		return False;
	}

	/* open for printing */

	cli_set_smb_cmd(cli, t_idx,  SMBsplopen, 2,2 + strlen(rname),True);

	SSVAL(cli->outbuf,smb_vwv0,0);
	SSVAL(cli->outbuf,smb_vwv1,info->print_mode);

	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p,rname);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL)) return False;

	fnum = SVAL(cli->inbuf,smb_vwv0);

	DEBUG(1,("cli_print: printing file %s as %s\n",lname,CNV_LANG(rname)));

	/* read file and send it */

	while (!feof(f))
	{
		int n;

		bzero(cli->outbuf,smb_size);
		set_message(cli->outbuf,1,3,True);

		/* for some strange reason the OS/2 print server can't handle large
		packets when printing. weird */
		n = MIN(1024, cli->max_xmit-(smb_len(cli->outbuf)+4));

		if (info->translation)
		{
			n = printread(info->translation, f,smb_buf(cli->outbuf)+3,(int)(0.95*n));
		}
		else
		{
			n = readfile(info->translation, smb_buf(cli->outbuf)+3,1,n,f);
		}

		if (n <= 0) 
		{
			DEBUG(0,("read gave %d\n",n));
			break;
		}

		smb_setlen(cli->outbuf,smb_len(cli->outbuf) + n);

		CVAL(cli->outbuf,smb_com) = SMBsplwr;
		SSVAL(cli->outbuf,smb_tid,cli->con[t_idx].cnum );
		cli_setup_packet(cli);

		SSVAL(cli->outbuf,smb_vwv0,fnum);
		SSVAL(cli->outbuf,smb_vwv1,n+3);
		CVAL(smb_buf(cli->outbuf),0) = 1;
		SSVAL(smb_buf(cli->outbuf),1,n);

		send_smb(cli->fd,cli->outbuf);
		if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
		if (cli_error(cli,  NULL, NULL))
		{
			DEBUG(0,("%s printing remote file\n", cli_errstr(cli)));
			break;
		}

		nread += n;
	}

	DEBUG(2,("%d bytes printed\n", nread));

	cli_set_smb_cmd(cli, t_idx,  SMBsplclose, 1,0,True);

	SSVAL(cli->outbuf,smb_vwv0,fnum);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL))
	{
		DEBUG(0,("%s closing print file\n", cli_errstr(cli)));
		if (f != stdin) fclose(f);
		return False;
	}

	if (f != stdin) fclose(f);

	return True;
}

/****************************************************************************
show a print queue - this is deprecated as it uses the old smb that
has limited support - the correct call is the cmd_p_queue_4() after this.
****************************************************************************/
int cli_queue(struct cli_state *cli, int t_idx, struct client_info *info,
				void (*fn)(uint16, char*, uint32, uint8))
{
	int count;
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBsplretq, 2,0,True);

	SSVAL(cli->outbuf, smb_vwv0, 32); /* a max of 20 entries is to be shown */
	SSVAL(cli->outbuf, smb_vwv1, 0); /* the index into the queue */

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL))
	{
		DEBUG(0,("%s obtaining print queue\n", cli_errstr(cli)));
		return -1;
	}

	count = SVAL(cli->inbuf,smb_vwv0);
	p = smb_buf(cli->inbuf) + 3;

	if (fn && count > 0)
	{
		int ct = count;

		while (ct--)
		{
			fn(SVAL(p,5), p+12, IVAL(p,7), CVAL(p, 4));
			p += 28;
		}
	}

	return count;
}

/****************************************************************************
  cancel a print job
  ****************************************************************************/
BOOL cli_cancel(struct cli_state *cli, int t_idx, uint16 job, uint16 *cancelled)
{
  BOOL ret = False;
	struct mem_buf rparam;
	struct mem_buf rdata ;
  char *p;
  pstring param;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

  bzero(param,sizeof(param));

  p = param;
  SSVAL(p,0,81);		/* DosPrintJobDel() */
  p += 2;
  strcpy(p,"W");
  p = skip_string(p,1);
  strcpy(p,"");
  p = skip_string(p,1);
  SSVAL(p,0,job);     
  p += 2;

	cli_send_trans(cli, t_idx, SMBtrans2,NULL,0,FID_UNUSED,0,
	         NULL,param,NULL,
	         0,PTR_DIFF(p,param),0,
	         BUFFER_SIZE,10,0);

	ret = cli_receive_trans(cli, t_idx,  SMBtrans2, &rdata,&rparam);

    if (ret)
    {
      if (cancelled) *cancelled = SVAL(rparam.data,0);
    }

	mem_free_data(&rparam);
	mem_free_data(&rdata );

  return ret;
}

/*****************************************************************************
 Convert a character pointer in a cli_call_api() response to a form we can use.
 This function contains code to prevent core dumps if the server returns 
 invalid data.
*****************************************************************************/
static char *fix_char_ptr(unsigned int datap, unsigned int converter,
				struct mem_buf *rdata)
{
if( datap == 0 )		/* turn NULL pointers */
  {				/* into zero length strings */
  return "";
  }
else
  {
  unsigned int offset = datap - converter;

  if( offset >= rdata->data_used )
    {
      DEBUG(1,("bad char ptr: datap=%u, converter=%u, rdata=%lu, rdata.data_used=%d>",
                datap, converter, (unsigned long)(rdata->data), rdata->data_used));
    return "<ERROR>";
    }
  else
    {
    return &(rdata->data)[offset];
    }
  }
}


/****************************************************************************
get information in a print queue
****************************************************************************/
int cli_pqueue_2(struct cli_state *cli, int t_idx, struct client_info *info,
			void (*fn)(char*, uint16, uint16, char *, time_t, uint32, char *))
{
	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	pstring param;
	int result_code = -1;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	bzero(param,sizeof(param));

	p = param;
	SSVAL(p,0,76);                            /* API fn 76 (DosPrintJobEnum) */
	p += 2;
	strcpy(p,"zWrLeh");                       /* parameter description? */
	p = skip_string(p,1);
	strcpy(p,"WWzWWDDzz");                    /* returned data format */
	p = skip_string(p,1);
	strcpy(p,strrchr(cli->con[t_idx].full_share,'\\')+1); /* name of queue */
	p = skip_string(p,1);
	SSVAL(p,0,2);                             /* API lvl2, PRJINFO_2 data */
	SSVAL(p,2,1000);                          /* size (bytes) of return buffer */
	p += 4;
	strcpy(p,"");                             /* subformat */
	p = skip_string(p,1);

	DEBUG(1,("Calling DosPrintJobEnum()...\n"));

	cli_send_trans(cli, t_idx, SMBtrans2,NULL,0,FID_UNUSED,0,
	         NULL,param,NULL,
	         0,PTR_DIFF(p,param),0,
	         BUFFER_SIZE,10,0);

	if (cli_receive_trans(cli, t_idx,  SMBtrans2,
	                       &rdata,&rparam))
	{
		int converter;
		int i;
		char PrinterName[20];

		result_code = SVAL(rparam.data,0);
		converter = SVAL(rparam.data,2);             /* conversion factor */

		DEBUG(2,("returned %d bytes of parameters, %d bytes of data, %d records\n", rparam.data_used, rdata.data_used, SVAL(rparam.data,4) ));

		if (result_code != 0) return result_code;

		strcpy(PrinterName,strrchr(cli->con[t_idx].full_share,'\\')+1); /* queue name */
		strlower(PrinterName);                              /* in lower case */

		p = rdata.data; /* received data */

		for (i = 0; i < SVAL(rparam.data,4); ++i)
		{
			uint16 JobId = SVAL(p,0);
			uint16 Priority = SVAL(p,2);
			char *UserName = fix_char_ptr(SVAL(p,4), converter, &rdata);
			time_t JobTime = make_unix_date3( p + 12);
			uint32 Size = IVAL(p,16);
			char *JobName = fix_char_ptr(SVAL(p,24), converter, &rdata);

			strlower(UserName);

			if (fn)
			{
				fn( PrinterName, JobId, Priority,
				    UserName,JobTime, Size, JobName);
			}

#if 0 /* DEBUG code */
			DEBUG(4,("Job Id: \"%u\"\n", SVAL(p,0)));
			DEBUG(4,("Priority: \"%u\"\n", SVAL(p,2)));

			DEBUG(4,("User Name: \"%s\"\n", fix_char_ptr(SVAL(p,4), converter, &rdata, rdata.data_used) ));
			DEBUG(4,("Position: \"%u\"\n", SVAL(p,8)));
			DEBUG(4,("Status: \"%u\"\n", SVAL(p,10)));

			JobTime = make_unix_date3( p + 12);
			DEBUG(4,("Submitted: \"%s\"\n", asctime(LocalTime(&JobTime))));
			DEBUG(4,("date: \"%u\"\n", SVAL(p,12)));

			DEBUG(4,("Size: \"%u\"\n", SVAL(p,16)));
			DEBUG(4,("Comment: \"%s\"\n", fix_char_ptr(SVAL(p,20), converter, &rdata, rdata.data_used) ));
			DEBUG(4,("Document: \"%s\"\n", fix_char_ptr(SVAL(p,24), converter, &rdata, rdata.data_used) ));
#endif /* DEBUG CODE */ 

			p += 28;
		}
	}
	else                  /* cli_call_api() failed */
	{
		DEBUG(1, ("Failed, error = %d\n", result_code));
	}

	/* If any parameters or data were returned, free the storage. */
	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return result_code;
}

/****************************************************************************
show information about a print queue
****************************************************************************/
BOOL cli_printq_info(struct cli_state *cli, int t_idx, struct client_info *info,
				char *name, uint16 *priority,
				uint16 *start_time, uint16 *until_time,
				char *separator_file, char *print_processor,
				char *params, char *comment,
				uint16 *status, uint16 *jobs,
				char *printers, char *driver_name,
				char **driver_data, int *driver_count)
{
	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	pstring param;
	int result_code = -1;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	bzero(param,sizeof(param));

	p = param;
	SSVAL(p,0,70); 			/* API function number 70 (DosPrintQGetInfo) */
	p += 2;
	strcpy(p,"zWrLh");			/* parameter description? */
	p = skip_string(p,1);
	strcpy(p,"zWWWWzzzzWWzzl");		/* returned data format */
	p = skip_string(p,1);
	strcpy(p,strrchr(cli->con[t_idx].full_share,'\\')+1);	/* name of queue */
	p = skip_string(p,1);
	SSVAL(p,0,3);				/* API function level 3, just queue info, no job info */
	SSVAL(p,2,1000);			/* size of bytes of returned data buffer */
	p += 4;
	strcpy(p,"");				/* subformat */
	p = skip_string(p,1);

	DEBUG(1,("Calling DosPrintQueueGetInfo()...\n"));

	cli_send_trans(cli, t_idx, SMBtrans2,NULL,0,FID_UNUSED,0,
			 NULL,param, NULL,
			 0,PTR_DIFF(p,param),1,
			 BUFFER_SIZE,10,0);

	if (cli_receive_trans(cli, t_idx,  SMBtrans2, &rparam, &rdata) )
	{
		int converter;
		result_code = SVAL(rparam.data,0);
		converter = SVAL(rparam.data,2);		/* conversion factor */

		DEBUG(2,("returned %d bytes of parameters, %d bytes of data, %d records\n", rparam.data_used, rdata.data_used, SVAL(rparam.data,4) ));

		if (result_code != 0) return result_code;

		p = rdata.data;				/* received data */

		fstrcpy(name, fix_char_ptr(SVAL(p,0), converter, &rdata) );
		*priority = SVAL(p,4);
		*start_time = SVAL(p,6);
		*until_time = SVAL(p,8);
		fstrcpy(separator_file, fix_char_ptr(SVAL(p,12), converter, &rdata) );
		fstrcpy(print_processor, fix_char_ptr(SVAL(p,16), converter, &rdata) );
		fstrcpy(params, fix_char_ptr(SVAL(p,20), converter, &rdata) );
		fstrcpy(comment, fix_char_ptr(SVAL(p,24), converter, &rdata) );
		*status = SVAL(p,28);
		*jobs = SVAL(p,30);
		fstrcpy(printers, fix_char_ptr(SVAL(p,32), converter, &rdata) );
		fstrcpy(driver_name, fix_char_ptr(SVAL(p,36), converter, &rdata) );

		(*driver_data) = rdata.data + SVAL(p,40) - converter;
		if( SVAL(p,40) == 0 )
		{
			*driver_count = 0;
		}
		else
		{
			*driver_count = IVAL((*driver_data),0);
		}
	}

	/* If any parameters or data were returned, free the storage. */
	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return result_code;
}

