/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client generic functions
   Copyright (C) Andrew Tridgell              1994-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef SYSLOG
#undef SYSLOG
#endif

#include "includes.h"
#include "trans2.h"

/* value for unused fid field in trans2 secondary request */
#define FID_UNUSED (0xFFFF)

#define CNV_LANG(s) dos2unix_format(s,False)

extern file_info def_finfo;

extern int DEBUGLEVEL;

/****************************************************************************
setup basics in a outgoing packet
****************************************************************************/
void cli_setup_packet(struct cli_state *cli)
{
	SSVAL(cli->outbuf,smb_pid,cli->pid);
	SSVAL(cli->outbuf,smb_uid,cli->uid);
	SSVAL(cli->outbuf,smb_mid,cli->mid);
	if (cli->protocol > PROTOCOL_CORE) {
		SCVAL(cli->outbuf,smb_flg,0x8);
		SSVAL(cli->outbuf,smb_flg2,0x1);
	}
}


/****************************************************************************
setup basics in a outgoing packet
****************************************************************************/
void cli_set_smb_cmd(struct cli_state *cli, int t_idx,
				uint8 cmd, int num_wds, int num_bytes, BOOL zero)
{
	if (!cli->inbuf || !cli->outbuf)
	{
		DEBUG(0,("cli_set_smb_cmd: buffers not allocated\n"));
		return;
	}

	bzero(cli->outbuf,smb_size);
	bzero(cli->inbuf ,smb_size);

	set_message(cli->outbuf, num_wds, num_bytes, zero);

	CVAL(cli->outbuf,smb_com) = cmd;
	SSVAL(cli->outbuf, smb_tid, cli->con[t_idx].cnum);

	cli_setup_packet(cli);
}


static  struct {
    int prot;
    char *name;
  }
prots[] = 
    {
      {PROTOCOL_CORE,"PC NETWORK PROGRAM 1.0"},
      {PROTOCOL_COREPLUS,"MICROSOFT NETWORKS 1.03"},
      {PROTOCOL_LANMAN1,"MICROSOFT NETWORKS 3.0"},
      {PROTOCOL_LANMAN1,"LANMAN1.0"},
      {PROTOCOL_LANMAN2,"LM1.2X002"},
      {PROTOCOL_LANMAN2,"Samba"},
      {PROTOCOL_NT1,"NT LANMAN 1.0"},
      {PROTOCOL_NT1,"NT LM 0.12"},
      {-1,NULL}
    };



/****************************************************************************
send a session setup

password-generating options:

- send two 16 byte hashes             (pass=LM16 ntpass=NT16)
- send a clear-text password          (pass=ct-pw ntpass=NULL)
- send different clear-text passwords (pass=ct-pw ntpass=ct-pw)
- send pre-computed OWFS              (pass=p24 ntpass=p24)

in all instances except the pre-computed OWFs, both the LM OWF and NT OWF
will be generated.

****************************************************************************/
BOOL cli_session_setup(struct cli_state *cli,
		       char *user, 
		       char *pass, int passlen,
		       uchar *ntpass, int ntpasslen,
		       char *workgroup)
{
	char *p;
	fstring pword  ;
	fstring ntpword;

	DEBUG(3,("cli_session_setup: protocol %d sec_mode: %x passlen %d\n",
	          cli->protocol, cli->sec_mode, passlen));

	if (cli->num_tcons != 0)
	{
		DEBUG(2,("cli_session_setup: connections appear to be already open\n"));
		return False;
	}

#ifdef DEBUG_PASSWORD
	if (pass)
	{
		DEBUG(100, ("password   : "));
		dump_data(100, pass, passlen);
	}
	if (ntpass)
	{
		DEBUG(100, ("nt-password: "));
		dump_data(100, ntpass, ntpasslen);
	}
#endif

	if (cli->protocol < PROTOCOL_LANMAN1)
	{
		DEBUG(5,("cli_session_setup: protocol < LANMAN1: sesssetup not needed\n"));
		return True;
	}

	if (passlen > sizeof(pword)-1) {
		return False;
	}

	if (IS_BITS_SET_ALL(cli->sec_mode, USE_CHALLENGE_RESPONSE))
	{
		DEBUG(5,("cli_session_setup: using challenge-response mode\n"));

		/* detect if two 16 byte hashes have been handed to us */
		if (pass && ntpass && passlen == 16 && ntpasslen == 16)
		{
			DEBUG(6,("cli_session_setup: OWF both lm and nt passwords\n"));

			passlen = 24;
			SMBOWFencrypt(pass  , (uchar *)cli->cryptkey, (uchar *)pword);
			ntpasslen = 24;
			SMBOWFencrypt(ntpass, (uchar *)cli->cryptkey, (uchar *)ntpword);
		}
		else if (pass)
		{
			if (*pass && passlen != 24)
			{
				/* do a LM password encrypt */
				passlen = 24;
				SMBencrypt(pass,(uchar *)cli->cryptkey,(uchar *)pword);

				if (!ntpass)
				{
					/* do an NT password encrypt */
					ntpasslen = 24;
					SMBNTencrypt(pass,(uchar *)cli->cryptkey,(uchar *)ntpword);
				}
			}
			else
			{
				memcpy(pword, pass, passlen);
			}

			if (ntpass)
			{
				if (*ntpass && ntpasslen != 24)
				{
					ntpasslen = 24;
					SMBNTencrypt((uchar *)ntpass,(uchar *)cli->cryptkey,(uchar *)ntpword);
				}
				else
				{
					/* an already-OWF'd challenge has been handed to us */
					memcpy(ntpword, ntpass, ntpasslen);
				}
			}
		}
		else
		{
			/* blank password... */
			pword[0] = 0;
			passlen = 1;
		}
	}
	else
	{
		DEBUG(5,("cli_session_setup: using given password (probably pass-through mode)\n"));
		memcpy(pword  , pass  , passlen  );
		memcpy(ntpword, ntpass, ntpasslen);
	}

	/* if in share level security then don't send a password now */
	if (!IS_BITS_SET_ALL(cli->sec_mode, USE_USER_LEVEL_SECURITY))
	{
		DEBUG(5,("cli_session_setup: using share-level security mode\n"));
		pword[0] = 0;
		passlen=1;
	} 

	/* send a session setup command */
	bzero(cli->outbuf,smb_size);

	if (cli->protocol < PROTOCOL_NT1)
	{
		set_message(cli->outbuf,10,1 + strlen(user) + passlen,True);
		CVAL(cli->outbuf,smb_com) = SMBsesssetupX;
		cli_setup_packet(cli);

		CVAL(cli->outbuf,smb_vwv0) = 0xFF;
		SSVAL(cli->outbuf,smb_vwv2,cli->max_xmit);
		SSVAL(cli->outbuf,smb_vwv3,2);
		SSVAL(cli->outbuf,smb_vwv4,1);
		SIVAL(cli->outbuf,smb_vwv5,cli->sesskey);
		SSVAL(cli->outbuf,smb_vwv7,passlen);
		p = smb_buf(cli->outbuf);
		memcpy(p,pword,passlen);
		p += passlen;
		strcpy(p,user);
		strupper(p);
	}
	else
	{
		cli_setpid(cli, 0xcafe);
		set_message(cli->outbuf,13,0,True);
		CVAL(cli->outbuf,smb_com) = SMBsesssetupX;
		cli_setup_packet(cli);
		
		CVAL(cli->outbuf,smb_vwv0) = 0xFF;
		SSVAL(cli->outbuf,smb_vwv2,BUFFER_SIZE);
		SSVAL(cli->outbuf,smb_vwv3,2);
		SSVAL(cli->outbuf,smb_vwv4,1);
		SIVAL(cli->outbuf,smb_vwv5,cli->sesskey);
		SSVAL(cli->outbuf,smb_vwv7,passlen);
		SSVAL(cli->outbuf,smb_vwv8,ntpasslen);
		SSVAL(cli->outbuf,smb_vwv11,CAP_STATUS32);
		p = smb_buf(cli->outbuf);
		memcpy(p,pword,passlen); 
		p += SVAL(cli->outbuf,smb_vwv7);
		if (ntpasslen != 0)
		{
			memcpy(p,ntpword,ntpasslen); 
			p += SVAL(cli->outbuf,smb_vwv8);
		}
		strcpy(p,user);
#if 0
		strupper(p);
#endif
		p = skip_string(p,1);
		strcpy(p,workgroup);
		strupper(p);
		p = skip_string(p,1);
		strcpy(p,"Windows NT 1381");p = skip_string(p,1);
		strcpy(p,"Windows NT 4.0");p = skip_string(p,1);
		set_message(cli->outbuf,13,PTR_DIFF(p,smb_buf(cli->outbuf)),False);
	}

      show_msg(cli->outbuf);

      send_smb(cli->fd,cli->outbuf);
      if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
	      return False;

      show_msg(cli->inbuf);

      if (cli_error(cli, NULL, NULL)) return(False);

      /* use the returned uid from now on */
      cli->uid = SVAL(cli->inbuf,smb_uid);

      return True;
}


/****************************************************************************
send a tconX
****************************************************************************/
BOOL cli_send_tconX(struct cli_state *cli, int *t_idx, 
		    char *share, char *dev, char *pass, int passlen)
{
	fstring pword;
	char *p;
	bzero(cli->outbuf,smb_size);
	bzero(cli->inbuf,smb_size);

	DEBUG(3,("cli_send_tconX: share:%s dev:%s passlen:%d\n",
			share, dev, passlen));

	if (cli->num_tcons == MAX_CLIENT_CONNECTIONS)
	{
		DEBUG(2,("cli_send_tconX: max number of client connections reached\n"));
		return False;
	}

	(*t_idx) = cli->num_tcons;
	cli->num_tcons++;

	if (!IS_BITS_SET_ALL(cli->sec_mode, USE_USER_LEVEL_SECURITY))
	{
		DEBUG(5,("cli_send_tconX: using share-level security mode\n"));
		passlen = 1;
		pass = "";
	}

	if (IS_BITS_SET_ALL(cli->sec_mode, USE_CHALLENGE_RESPONSE) && *pass && passlen != 24)
	{
		DEBUG(5,("cli_send_tconX: using challenge-response mode\n"));
		passlen = 24;
		SMBencrypt((uchar *)pass,(uchar *)cli->cryptkey,(uchar *)pword);
	}
	else
	{
		DEBUG(5,("cli_send_tconX: using given password (probably pass-through mode)\n"));
		memcpy(pword, pass, passlen);
	}

	strcpy (cli->con[(*t_idx)].dev  , dev);
	fstrcpy(cli->con[(*t_idx)].share, share);
	sprintf(cli->con[(*t_idx)].full_share, "\\\\%s\\%s", cli->called.name, share);

	DEBUG(5,("cli_send_tconX: full share name:%s\n", cli->con[(*t_idx)].full_share));

	set_message(cli->outbuf,4,
		    2 + strlen(cli->con[(*t_idx)].full_share) + passlen + strlen(cli->con[(*t_idx)].dev),True);
	CVAL(cli->outbuf,smb_com) = SMBtconX;
	cli_setup_packet(cli);

	SSVAL(cli->outbuf,smb_vwv0,0xFF);
	SSVAL(cli->outbuf,smb_vwv3,passlen);

	p = smb_buf(cli->outbuf);
	memcpy(p,pword,passlen);
	p += passlen;
	strcpy(p,cli->con[(*t_idx)].full_share);
	p = skip_string(p,1);
	strcpy(p,cli->con[(*t_idx)].dev);

	SCVAL(cli->inbuf,smb_rcls, 1);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
		return False;

    if (cli_error(cli, NULL, NULL)) return(False);

	cli->con[(*t_idx)].cnum = SVAL(cli->inbuf,smb_tid);
	return True;
}


/****************************************************************************
send a tree disconnect
****************************************************************************/
BOOL cli_tdis(struct cli_state *cli, int t_idx)
{
	cli_set_smb_cmd(cli, t_idx,  SMBtdis, 0,0,True);
	
	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
		return False;
	
    return !cli_error(cli, NULL, NULL);
}


/****************************************************************************
send a negprot command
****************************************************************************/
BOOL cli_negprot(struct cli_state *cli)
{
	char *p;
	int numprots;
	int plength;

	bzero(cli->outbuf,smb_size);

	DEBUG(3,("cli_negprot\n"));

	if (cli->num_tcons != 0)
	{
		DEBUG(2,("cli_negprot: connections appear to be already open\n"));
		return False;
	}

	/* setup the protocol strings */
	for (plength=0,numprots=0;
	     prots[numprots].name && prots[numprots].prot<=cli->protocol;
	     numprots++)
		plength += strlen(prots[numprots].name)+2;
    
	set_message(cli->outbuf,0,plength,True);

	p = smb_buf(cli->outbuf);
	for (numprots=0;
	     prots[numprots].name && prots[numprots].prot<=cli->protocol;
	     numprots++) {
		*p++ = 2;
		strcpy(p,prots[numprots].name);
		p += strlen(p) + 1;
	}

	CVAL(cli->outbuf,smb_com) = SMBnegprot;
	cli_setup_packet(cli);

	CVAL(smb_buf(cli->outbuf),0) = 2;

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
		return False;

	show_msg(cli->inbuf);

    if (cli_error(cli, NULL, NULL)) return False;
	if ((int)SVAL(cli->inbuf,smb_vwv0) >= numprots) return(False);

	cli->protocol = prots[SVAL(cli->inbuf,smb_vwv0)].prot;


	if (cli->protocol >= PROTOCOL_NT1) {    
		/* NT protocol */
		cli->sec_mode = CVAL(cli->inbuf,smb_vwv1);
		cli->max_xmit = IVAL(cli->inbuf,smb_vwv3+1);
		cli->sesskey = IVAL(cli->inbuf,smb_vwv7+1);
		cli->serverzone = SVALS(cli->inbuf,smb_vwv15+1)*60;
		/* this time arrives in real GMT */
		cli->servertime = interpret_long_date(cli->inbuf+smb_vwv11+1);
		memcpy(cli->cryptkey,smb_buf(cli->inbuf),8);
		if (IVAL(cli->inbuf,smb_vwv9+1) & 1)
			cli->readbraw_supported = 
				cli->writebraw_supported = True;      
	} else if (cli->protocol >= PROTOCOL_LANMAN1) {
		cli->sec_mode = SVAL(cli->inbuf,smb_vwv1);
		cli->max_xmit = SVAL(cli->inbuf,smb_vwv2);
		cli->sesskey = IVAL(cli->inbuf,smb_vwv6);
		cli->serverzone = SVALS(cli->inbuf,smb_vwv10)*60;
		/* this time is converted to GMT by make_unix_date */
		cli->servertime = make_unix_date(cli->inbuf+smb_vwv8);
		cli->readbraw_supported = ((SVAL(cli->inbuf,smb_vwv5) & 0x1) != 0);
		cli->writebraw_supported = ((SVAL(cli->inbuf,smb_vwv5) & 0x2) != 0);
		memcpy(cli->cryptkey,smb_buf(cli->inbuf),8);
	} else {
		/* the old core protocol */
		cli->sec_mode = 0;
		cli->serverzone = TimeDiff(time(NULL));
	}

	DEBUG(5,("cli_negprot: secmode:%x\n", cli->sec_mode));

	return True;
}


/****************************************************************************
  send a session request.  see rfc1002.txt 4.3 and 4.3.2
****************************************************************************/
BOOL cli_session_request(struct cli_state *cli,
			struct nmb_name *calling, struct nmb_name *called)
{
	char *p;
	int len = 4;
	/* send a session request (RFC 1002) */

	memcpy(&(cli->calling), calling, sizeof(*calling));
	memcpy(&(cli->called ), called , sizeof(*called ));
  
	/* put in the destination name */
	p = cli->outbuf+len;
	name_mangle(cli->called .name, p, cli->called .name_type);
	len += name_len(p);

	/* and my name */
	p = cli->outbuf+len;
	name_mangle(cli->calling.name, p, cli->calling.name_type);
	len += name_len(p);

	/* setup the packet length */
	_smb_setlen(cli->outbuf,len);
	CVAL(cli->outbuf,0) = 0x81;

	send_smb(cli->fd,cli->outbuf);
	DEBUG(5,("Sent session request\n"));

	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
		return False;

	if (CVAL(cli->inbuf,0) != 0x82) {
		cli->error = CVAL(cli->inbuf,0);
		return False;
	}
	return(True);
}


/****************************************************************************
open the client sockets
****************************************************************************/
BOOL cli_connect(struct cli_state *cli, char *host, struct in_addr *ip)
{
	fstrcpy(cli->full_dest_host_name, host);
	
	if (!ip || zero_ip(*ip))
	{
		/* no ip specified - look up the name */
		struct hostent *hp;

		if ((hp = Get_Hostbyname(host)) == 0) {
			return False;
		}

		putip((char *)&(cli->dest_ip),(char *)hp->h_addr);
	} else {
		/* use the given ip address */
		cli->dest_ip = *ip;
	}

	/* open the socket */
	cli->fd = open_socket_out(SOCK_STREAM, &(cli->dest_ip), 139, cli->timeout);

	return (cli->fd != -1);
}


/****************************************************************************
initialise a client structure
****************************************************************************/
BOOL cli_initialise(struct cli_state *cli)
{
	int i;

	DEBUG(5,("cli_initialise\n"));

	if (cli->initialised) cli_shutdown(cli);

	memset(cli, 0, sizeof(*cli));
	cli->fd = -1;

	for (i = 0; i < MAX_CLIENT_CONNECTIONS; i++)
	{
		cli->con[i].cnum = -1;
	}
	cli->num_tcons = 0;

	cli->pid = getpid();
	cli->mid = 1;
	cli->uid = getuid();
	cli->protocol = PROTOCOL_NT1;
	cli->timeout = 20000;
	cli->bufsize = 0x10000;
	cli->max_xmit = cli->bufsize - 4;
	cli->outbuf = (char *)malloc(cli->bufsize);
	cli->inbuf  = (char *)malloc(cli->bufsize);

	if (!cli->outbuf || !cli->inbuf)
	{
		DEBUG(5,("cli_initialise: buffer allocation failed\n"));
		return False;
	}

	cli->initialised = 1;

	DEBUG(5,("cli_initialise: ok\n"));

	return True;
}

/****************************************************************************
shutdown a client structure
****************************************************************************/
void cli_shutdown(struct cli_state *cli)
{
	DEBUG(5,("cli_shutdown\n"));

	if (cli->outbuf) free(cli->outbuf);
	if (cli->inbuf) free(cli->inbuf);
	if (cli->fd != -1) close(cli->fd);
	memset(cli, 0, sizeof(*cli));
}

/****************************************************************************
  return a description of the error
****************************************************************************/
char *cli_errstr(struct cli_state *cli)
{
	return smb_errstr(cli->inbuf);
}

/****************************************************************************
  return error codes for the last packet
****************************************************************************/
BOOL cli_error(struct cli_state *cli, uint8 *eclass, uint32 *num)
{
	int  flgs2 = SVAL(cli->inbuf,smb_flg2);

	if (eclass) *eclass = 0;
	if (num   ) *num = 0;

	if (flgs2 & FLAGS2_32_BIT_ERROR_CODES)
	{
		/* 32 bit error codes detected */
		uint32 nt_err = IVAL(cli->inbuf,smb_rcls);
		if (num) *num = nt_err;
		DEBUG(10,("cli_error: 32 bit codes: code=%08x\n", nt_err));
		return (IS_BITS_SET_ALL(nt_err, 0xc0000000));
	}
	else
	{
		/* dos 16 bit error codes detected */
		char rcls  = CVAL(cli->inbuf,smb_rcls);
		if (rcls != 0)
		{
			if (eclass) *eclass = rcls;
			if (num   ) *num    = SVAL(cli->inbuf,smb_err);
			return True;
		}
	}
	return False;
}

/****************************************************************************
set socket options on a open connection
****************************************************************************/
void cli_sockopt(struct cli_state *cli, char *options)
{
	set_socket_options(cli->fd, options);
}

/****************************************************************************
set the MID to use for smb messages. Return the old MID.
****************************************************************************/
int cli_setmid(struct cli_state *cli, int mid)
{
	int ret = cli->mid;
	cli->mid = mid;
	return ret;
}

/****************************************************************************
set the PID to use for smb messages. Return the old pid.
****************************************************************************/
int cli_setpid(struct cli_state *cli, int pid)
{
	int ret = cli->pid;
	cli->pid = pid;
	return ret;
}


/****************************************************************************
establishes a connection right up to doing tconX, reading in a password.
****************************************************************************/
BOOL cli_reestablish_connection(struct cli_state *cli, int *t_idx)
{
	struct nmb_name calling;
	struct nmb_name called;
	fstring dest_host;
	struct in_addr dest_ip;
	fstring share;
	fstring dev;
	BOOL do_tcon = False;

	if (!cli->initialised || cli->fd == -1)
	{
		DEBUG(3,("cli_reestablish_connection: not connected\n"));
		return False;
	}

	/* copy the parameters necessary to re-establish the connection */

	if ((*t_idx) != -1)
	{
		fstrcpy(share, cli->con[*t_idx].share);
		fstrcpy(dev  , cli->con[*t_idx].dev);
		do_tcon = True;
	}

	memcpy(&called , &(cli->called ), sizeof(called ));
	memcpy(&calling, &(cli->calling), sizeof(calling));
	fstrcpy(dest_host, cli->full_dest_host_name);
	dest_ip = cli->dest_ip;

	DEBUG(5,("cli_reestablish_connection: %s connecting to %s (ip %s) - %s [%s]\n",
		          namestr(&calling), namestr(&called), inet_ntoa(dest_ip),
	              cli->user_name, cli->domain));

	return cli_establish_connection(cli, t_idx,
	                                dest_host, &dest_ip,
	                                &calling, &called,
	                                share, dev, False, do_tcon);
}

/****************************************************************************
establishes a connection right up to doing tconX, reading in a password.
****************************************************************************/
BOOL cli_establish_connection(struct cli_state *cli, int *t_idx,
				char *dest_host, struct in_addr *dest_ip,
				struct nmb_name *calling, struct nmb_name *called,
				char *service, char *service_type,
				BOOL do_shutdown, BOOL do_tcon)
{
	DEBUG(5,("cli_establish_connection: %s connecting to %s (%s) - %s [%s]\n",
		          namestr(calling), namestr(called), inet_ntoa(*dest_ip),
	              cli->user_name, cli->domain));

	/* establish connection */

	if ((!cli->initialised))
	{
		return False;
	}

	if (cli->fd == -1)
	{
		if (!cli_connect(cli, dest_host, dest_ip))
		{
			DEBUG(1,("cli_establish_connection: failed to connect to %s (%s)\n",
					  namestr(calling), inet_ntoa(*dest_ip)));
			return False;
		}
	}

	if (!cli_session_request(cli, calling, called))
	{
		DEBUG(1,("failed session request\n"));
		if (do_shutdown) cli_shutdown(cli);
		return False;
	}

	if (!cli_negprot(cli))
	{
		DEBUG(1,("failed negprot\n"));
		if (do_shutdown) cli_shutdown(cli);
		return False;
	}

	if (cli->pwd.cleartext || cli->pwd.null_pwd)
	{
		/* attempt clear-text session */

		fstring passwd;

		pwd_get_cleartext(&(cli->pwd), passwd);

		/* attempt clear-text session */
		if (!cli_session_setup(cli, cli->user_name,
	                       passwd, strlen(passwd),
	                       NULL, 0,
	                       cli->domain))
		{
			DEBUG(1,("failed session setup\n"));
			if (do_shutdown) cli_shutdown(cli);
			return False;
		}
		if (do_tcon)
		{
			if (!cli_send_tconX(cli, t_idx,  service, service_type,
			                    (char*)passwd, strlen(passwd)))
			{
				DEBUG(1,("failed tcon_X\n"));
				if (do_shutdown) cli_shutdown(cli);
				return False;
			}
		}
	}
	else
	{
		/* attempt encrypted session */
		uchar nt_sess_pwd[24];
		uchar lm_sess_pwd[24];

		/* creates (storing a copy of) and then obtains a 24 byte password OWF */
		pwd_make_lm_nt_owf(&(cli->pwd), cli->cryptkey);
		pwd_get_lm_nt_owf(&(cli->pwd), lm_sess_pwd, nt_sess_pwd);

		/* attempt encrypted session */
		if (!cli_session_setup(cli, cli->user_name,
	                       (char*)lm_sess_pwd, sizeof(lm_sess_pwd),
	                              nt_sess_pwd, sizeof(nt_sess_pwd),
	                       cli->domain))
		{
			DEBUG(1,("failed session setup\n"));
			if (do_shutdown) cli_shutdown(cli);
			return False;
		}

		if (do_tcon)
		{
			if (!cli_send_tconX(cli, t_idx,  service, service_type,
			                    (char*)nt_sess_pwd, sizeof(nt_sess_pwd)))
			{
				DEBUG(1,("failed tcon_X\n"));
				if (do_shutdown) cli_shutdown(cli);
				return False;
			}
		}
	}

	if (do_shutdown) cli_shutdown(cli);

	return True;
}
